from __future__ import absolute_import


class EstimatorException(Exception):
    pass


class ModelException(Exception):
    pass


class IllegalArgument(Exception):
    pass


class PredictionException(Exception):
    pass


class BacktestException(Exception):
    pass
